@@ -5,13 +5,14 @@ extern crate proc_macro;
 
 use pgx_utils::{categorize_return_type, CategorizedType};
 use proc_macro2::{Ident, Span};
-use quote::{quote, quote_spanned};
+use quote::{quote, quote_spanned, ToTokens};
 use std::ops::Deref;
 use std::str::FromStr;
+use syn::punctuated::Punctuated;
 use syn::spanned::Spanned;
 use syn::{
     FnArg, ForeignItem, ForeignItemFn, Generics, ItemFn, ItemForeignMod, Pat, ReturnType,
-    Signature, Type, Visibility,
+    Signature, Token, Type, Visibility,
 };
 
 pub struct PgGuardRewriter();
@@ -113,7 +114,10 @@ impl PgGuardRewriter {
                 true,
             ),
 
-            CategorizedType::Tuple(_types) => (PgGuardRewriter::impl_tuple_udf(func), false),
+            CategorizedType::Tuple(_types) => (
+                PgGuardRewriter::impl_tuple_udf(func, entity_submission.clone()),
+                false,
+            ),
 
             CategorizedType::Iterator(types) if types.len() == 1 => (
                 PgGuardRewriter::impl_setof_srf(
@@ -211,20 +215,33 @@ impl PgGuardRewriter {
         }
     }
 
-    fn impl_tuple_udf(mut func: ItemFn) -> proc_macro2::TokenStream {
+    fn impl_tuple_udf(
+        mut func: ItemFn,
+        entity_submission: Option<&pgx_utils::sql_entity_graph::PgExtern>,
+    ) -> proc_macro2::TokenStream {
         let func_span = func.span();
         let return_type = func.sig.output;
         let return_type = format!("{}", quote! {#return_type});
         let return_type =
             proc_macro2::TokenStream::from_str(return_type.trim_start_matches("->")).unwrap();
         let return_type = quote! {impl std::iter::Iterator<Item = #return_type>};
+        let attrs = entity_submission
+            .unwrap()
+            .extern_attrs()
+            .iter()
+            .collect::<Punctuated<_, Token![,]>>();
 
         func.sig.output = ReturnType::Default;
         let sig = func.sig;
         let body = func.block;
+
         // We do **not** put an entity submission here as there still exists a `pg_extern` attribute.
+        //
+        // This is because we quietely rewrite the function signature to `Iterator<Item = T>` and
+        // rely on #[pg_extern] being called again during compilation.  It is important that we
+        // include the original #[pg_extern(<attributes>)] in the generated code.
         quote_spanned! {func_span=>
-            #[pg_extern]
+            #[pg_extern(#attrs)]
             #sig -> #return_type {
                 Some(#body).into_iter()
             }
@@ -696,7 +713,9 @@ impl FunctionSignatureRewriter {
     fn args(&self, is_raw: bool) -> proc_macro2::TokenStream {
         if self.func.sig.inputs.len() == 1 && self.return_type_is_datum() {
             if let FnArg::Typed(ty) = self.func.sig.inputs.first().unwrap() {
-                if type_matches(&ty.ty, "pg_sys :: FunctionCallInfo") {
+                if type_matches(&ty.ty, "pg_sys :: FunctionCallInfo")
+                    || type_matches(&ty.ty, "pgx :: pg_sys :: FunctionCallInfo")
+                {
                     return proc_macro2::TokenStream::new();
                 }
             }
@@ -704,7 +723,31 @@ impl FunctionSignatureRewriter {
 
         let mut stream = proc_macro2::TokenStream::new();
         let mut i = 0usize;
-        let mut have_fcinfo = false;
+        let mut fcinfo_ident = None;
+
+        // Get the fcinfo ident, if it exists.
+        // We do this because we need to get the **right** ident, if it exists, so Rustc
+        // doesn't think we're pointing at the fcinfo module path.
+        for arg in &self.func.sig.inputs {
+            match arg {
+                FnArg::Typed(ty) => match ty.pat.deref() {
+                    Pat::Ident(ident) => {
+                        if type_matches(&ty.ty, "pg_sys :: FunctionCallInfo")
+                            || type_matches(&ty.ty, "pgx :: pg_sys :: FunctionCallInfo")
+                        {
+                            if fcinfo_ident.is_some() {
+                                panic!("When using `pg_sys::FunctionCallInfo` as an argument it must be the last argument");
+                            }
+                            fcinfo_ident = Some(ident.ident.clone());
+                        }
+                    },
+                    _ => (),
+                },
+                _ => ()
+            }
+        }
+        let fcinfo_ident = fcinfo_ident.unwrap_or(syn::Ident::new("fcinfo", Span::call_site()));
+
         for arg in &self.func.sig.inputs {
             match arg {
                 FnArg::Receiver(_) => panic!("Functions that take self are not supported"),
@@ -714,31 +757,28 @@ impl FunctionSignatureRewriter {
                         let mut type_ = ty.ty.clone();
                         let is_option = type_matches(&type_, "Option");
 
-                        if have_fcinfo {
-                            panic!("When using `pg_sys::FunctionCallInfo` as an argument it must be the last argument")
-                        }
-
                         let ts = if is_option {
                             let option_type = extract_option_type(&type_);
                             let mut option_type = syn::parse2::<syn::Type>(option_type).unwrap();
                             pgx_utils::anonymonize_lifetimes(&mut option_type);
 
                             quote_spanned! {ident.span()=>
-                                let #name = pgx::pg_getarg::<#option_type>(fcinfo, #i);
+                                let #name = pgx::pg_getarg::<#option_type>(#fcinfo_ident, #i);
                             }
-                        } else if type_matches(&type_, "pg_sys :: FunctionCallInfo") {
-                            have_fcinfo = true;
+                        } else if type_matches(&type_, "pg_sys :: FunctionCallInfo")
+                            || type_matches(&type_, "pgx :: pg_sys :: FunctionCallInfo")
+                        {
                             quote_spanned! {ident.span()=>
-                                let #name = fcinfo;
+                                let #name = #fcinfo_ident;
                             }
                         } else if is_raw {
                             quote_spanned! {ident.span()=>
-                                let #name = pgx::pg_getarg_datum_raw(fcinfo, #i) as #type_;
+                                let #name = pgx::pg_getarg_datum_raw(#fcinfo_ident, #i) as #type_;
                             }
                         } else {
                             pgx_utils::anonymonize_lifetimes(&mut type_);
                             quote_spanned! {ident.span()=>
-                                let #name = pgx::pg_getarg::<#type_>(fcinfo, #i).unwrap_or_else(|| panic!("{} is null", stringify!{#ident}));
+                                let #name = pgx::pg_getarg::<#type_>(#fcinfo_ident, #i).unwrap_or_else(|| panic!("{} is null", stringify!{#ident}));
                             }
                         };
 
@@ -746,7 +786,10 @@ impl FunctionSignatureRewriter {
 
                         i += 1;
                     }
-                    _ => panic!("Unrecognized function arg type"),
+                    _ => panic!(
+                        "Unrecognized function arg type: {}",
+                        arg.to_token_stream().to_string()
+                    ),
                 },
             }
         }
